@@ -3,7 +3,7 @@
 
 extern void update_display (void);
 extern void configure (GtkWidget * widget, gpointer d);
-extern void edit_person (struct person *p, gboolean isnew);
+extern void edit_person (struct person *p, gchar *title, gboolean isdialog);
 extern void load_panel_config ();
 extern void update_edit (struct person *p, GtkWidget *w);
 extern gchar *build_categories_string (struct person *p);
